/*---------------------------------------------------------------------------\
| PIC18 CAN Peripheral Software for the PIC18F458
|
| File:					18F_CAN.C			Mainline code
|
| Global Routines:		CanInit()
|						CanTxIsr()
|						CanRxIsr()
|						SendCanMsg()
|						GetCanMsg()
|
| Local Routines:		StoreMsgInTXBn()
|						GetMsgFromRXBn() ????
|
| Discussion:        	None
|
| Overview:
|
|  Initialization:
|
|	Call InitializeCan to set everything up. See the enums for more details.
|
|  Transmission:
|
|	Mainline calls SendCanMsg which stuffs a message into the one of three
|	queues, CanTxQueue0, CanTxQueue1 or CanTxQueue2 based on the message's
|	ID. CanTx0Isr, CanTx1ISR, and CanTx2ISR then grab the messages from the
|	respective queues and stuff them into the CAN protocol unit's registers
|	to send them (TXB0, TXB1, TXB2). In other words:
|
|							  CanTx2Queue -> CanTx2ISR -> TXB2
|	Main -> SendCanMessage -> CanTx1Queue -> CanTx1ISR -> TXB1
|							  CanTx0Queue -> CanTx0ISR -> TXB0
|
|   We use three buffers because we have lots of RAM and the TXBn buffers
|	are sent in the order TXB2 to TXB0 - thus TXB0 could have very long
|	send latenacies if TXB2 and TXB1 are full. In other words, we
|	guarantee that high prioirty messages are sent to TXB2 and are always
| 	sent first, and that medium prioirty messages are sent as soon as
|	high prioirties are sent, etc.
|
|  Reception:
|
|	SerialRxIsr grabs bytes from the UART Receive register (RXREG) on a
|	Receiver Buffer Full Interrupt and stuffs them into the SerialRxQueue.
|	Mainline routines can then grab bytes from the queue using GetSerialByte.
|	In other words:
|
|	RCREG -> SerialRxISR -> SerialRxQueue -> GetSerialByte -> Main
|
|    Date    | Notes|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/28/2002 | Created.
\---------------------------------------------------------------------------*/


//------------------------------------------------------------------------
//  #includes and local #defines
//------------------------------------------------------------------------

#include <pic18.h>    	// Include for PICC-18
#include "pic18fxx8.h"	// Register definitions for the PIC18F458
#include "project.h"	// Includes for the project (Globals + misc)
#include "main.h"		// Includes for the mainline
#include "ints.h"		// Includes for interrupts
#include "excepts.h"	// Includes for exception handling
#include "can.h"		// Includes for the CAN peripheral



//------------------------------------------------------------------------
//  Global Variables (extern'ed in can.h)
//------------------------------------------------------------------------

CanMessage_t	canTx2Queue[CAN_TX2_QUEUE_SIZE];	// CAN Transmit queues.
uint8_t			canTx2IndexOut;
uint8_t			canTx2IndexIn;

CanMessage_t	canTx1Queue[CAN_TX1_QUEUE_SIZE];
uint8_t			canTx1IndexOut;
uint8_t			canTx1IndexIn;

CanMessage_t	canTx0Queue[CAN_TX0_QUEUE_SIZE];
uint8_t			canTx0IndexOut;
uint8_t			canTx0IndexIn;


//------------------------------------------------------------------------
//  Local function declarations
//------------------------------------------------------------------------

static void StoreCanMsgId	(uint8_t* pRegister, CanHeader_t value);
static void GetCanMsgId		(uint8_t* pRegister, CanHeader_t *value);



/*---------------------------------------------------------------------------\
| Function:		InitializeCan
|
| Purpose:		Initialize the CAN peripheral
|
| Discussion:	Pass the bitrate and the CAN masks and filters, and it
|				sets up the chip. Many hidden options which shouldn't
|				affect CAN operations hidden here and in SetCantiming.
|
| Expects:		-
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

 void InitializeCan (enum CAN_BITRATE_TYPE canBitRate,
                    CanHeader_t canMaskB0,
                    CanHeader_t canMaskB1,
                    CanHeader_t CanFilterB0F0,
                    CanHeader_t CanFilterB0F1,
                    CanHeader_t CanFilterB1F2,
                    CanHeader_t CanFilterB1F3,
                    CanHeader_t CanFilterB1F4,
                    CanHeader_t CanFilterB1F5 )
 {
	// Request config mode - INDETERMINATE TIME CALL!
	// COULD BE 100s of MS! After a reset, we're already
	// in Config Mode so it should take zero time.

    RequestCanMode(CAN_CONFIG_MODE);

    // Now set the CAN Bit rate.

	SetCanTiming(CAN_BITRATE_1MBPS);

	// Configure the receive regist. We're hard coding many options
	// since in our case, we'll never need it different (HA!).
	// In particular: Enable double buffer (RB0 ->RB1), receive
	// only valid CAN messages with a standard 11bit identifier,
	// and throw out any currently unread messages (shouldn't be
	// any after a reset).

	RXB0CON = 0b00100100;
		  //    ||||||||
		  //    |||||||+--- FILHIT0 = x
		  //    ||||||+---- JTOFF = x
		  //    |||||+----- RXB0DBEN = 1 -> Enable double buffer to RXB1
		  //    ||||+------ RXRTRRO = x
		  //    |||+------- <unimplemented on PIC>
		  //    ||+-------- RXM0 = 01 -> Receive only valid 11bit MsgIDs
		  //    |+--------- RXM1
	  	  //    +---------- RXFUL = 0 (throw out any unread message)

	RXB1CON = 0b00100000;
		  //    ||||||||
		  //    |||||||+--- FILHIT0 = x
		  //    ||||||+---- FILHIT1 = x
		  //    |||||+----- FILHIT2 = x
		  //    ||||+------ RXRTRRO = x
		  //    |||+------- <unimplemented on PIC>
		  //    ||+-------- RXM0 = 01 -> Receive only valid 11bit MsgIDs
		  //    |+--------- RXM1
	  	  //    +---------- RXFUL = 0 (throw out any unread message)

    // Set the receive masks and filters.

    CanSetMask(CAN_RX_MASK_B0, canMaskB0);
    CanSetMask(CAN_RX_MASK_B1, canMaskB1);

    CanSetFilter(CAN_RX_FILTER_B0_F0, CanFilterB0F0);
    CanSetFilter(CAN_RX_FILTER_B0_F1, CanFilterB0F1);
    CanSetFilter(CAN_RX_FILTER_B1_F2, CanFilterB1F2);
    CanSetFilter(CAN_RX_FILTER_B1_F3, CanFilterB1F3);
    CanSetFilter(CAN_RX_FILTER_B1_F4, CanFilterB1F4);
    CanSetFilter(CAN_RX_FILTER_B1_F5, CanFilterB1F5);

	// There's nothing to configure (now) for the transmit side -
	// that's all done when we send a message.

    // Restore to Normal mode.

    RequestCanMode(CAN_NORMAL_MODE);
}







/*---------------------------------------------------------------------------\
| Function:		SetCanTiming
|
| Purpose:		Set the bit rate and timing registers.
|
| Discussion:	Yikes, CAN bus timing is complicated. Here's were we stuff
|				the complexity in all of its glory.
|               Note that we also choose to use the CAN bus line filter
|				for wakeups, choose the # of times to sample, and the SJW
|				here... all of which are hardware-like things that shouldn't
|				need to be changed ever (hah!).
|
| Expects:		To be in configuration mode (else registers will not change
|				their value).
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/


void SetCanTiming (enum CAN_BITRATE_TYPE canBitRate)
{
	switch (canBitRate)
	{
		case (CAN_BITRATE_125KBPS):

			// 125kbps per second -> Tbit = 8us
			// ASSUME 20MHz CLOCK.
			// Divide by 10 to give us 16 CAN clocks (Tq's) in 1 bit.
			// Meaning 1Tq = 500ns.
			//
			//	|---|---|---|---|---|---|---|---|---|---|---|---|---|---|---|---|
			//	|   |Prop   | Seg1                      |Seg2                       |
			//  Sync                                    ^Sample = 62.5%
			//
			// which translates to a SEG1 = 7Tq and SEG2 = 6T if the Sync is 1Tq
			// and PROP is 2Tq.

			// The values to set the registers are really weird, so we'll set
			// them explicitly here and not try and hide the weirdness:

			BRGCON1 = 0b00000100;
				  //    ||||||||
				  //    |||||||+--- BRP0 = 000100 = Tq = 2*5/FOSC = 500ns -> 16Tq/bit
				  //    ||||||+---- BRP1
				  //    |||||+----- BRP2
				  //    ||||+------ BRP3
				  //    |||+------- BRP4
				  //    ||+-------- BRP5
				  //    |+--------- SJW = 00 -> SJW = 1Tq
				  //    +---------- SJW1

			BRGCON2 = 0b11110001;
				  //    ||||||||
				  //    |||||||+--- PRSEG0 = 001 	-> PROP = 2Tq
				  //    ||||||+---- PRSEG1
				  //    |||||+----- PRSEG2
				  //    ||||+------ SEG1PH0 = 110 	-> SEG1 = 7Tq
				  //    |||+------- SEG1PH1
				  //    ||+-------- SEG1PH2
				  //    |+--------- SAM = 1 		-> three samples
				  //    +---------- SEG2PHTS = 1 	-> Use SEG2PH bits

			BRGCON3 = 0b10100010;
				  //    ||||||||
				  //    |||||||+--- <Unimplemented on PIC>
				  //    ||||||+---- WAKFIL = 1 		-> Use CAN filter on wakeup
				  //    |||||+----- <Unimplemented on PIC>
				  //    ||||+------ <Unimplemented on PIC>
				  //    |||+------- <Unimplemented on PIC>
				  //    ||+-------- SEG2PH0 = 101 	-> SEG2 = 6Tq
				  //    |+--------- SEG2PH1
				  //    +---------- SEG2PH2

			break;

		case (CAN_BITRATE_1MBPS):
		default:

			// 1Mb per second -> Tbit = 1us
			// ASSUME 20MHz CLOCK.
			// Divide by 2 to give us 10 CAN clocks (Tq's) in 1 bit.
			// Meaning 1Tq = 100ns.
			//
			// IN LV2's case, the Flight Computer (Intel 82527) does 8Tq/bit:
			//
			//  |--------|--------|--------|--------|-------|--------|--------|-------|
			//	|Tsync   |Prop + Seg1                       |Seg2                     |
			//                                              ^Sample = 62.5%
			//
			// And the closest match we can do it at 10Tq/bit is:
			//
			//	|------|------|------|------|------|------|------|------|------|------|
			//	|Tsync |Prop  |Seg1                       |Seg2                       |
			//                                            ^Sample = 60.0%
			//
			// which translates to a SEG1 = 4Tq and SEG2 = 4T if the PROP and Tsync
			// segments are both 1Tq.

			// The values to set the registers are really weird, so we'll set
			// them explicitly here and not try and hide the weirdness:

			BRGCON1 = 0b00000000;
				  //    ||||||||
				  //    |||||||+--- BRP0 = 000000 = Tq = 2/FOSC = 100ns -> 10Tq/bit
				  //    ||||||+---- BRP1
				  //    |||||+----- BRP2
				  //    ||||+------ BRP3
				  //    |||+------- BRP4
				  //    ||+-------- BRP5
				  //    |+--------- SJW = 00 -> SJW = 1Tq
				  //    +---------- SJW1

			BRGCON2 = 0b10011000;
				  //    ||||||||
				  //    |||||||+--- PRSEG0 = 000 	-> PROP = 1Tq
				  //    ||||||+---- PRSEG1
				  //    |||||+----- PRSEG2
				  //    ||||+------ SEG1PH0 = 011 	-> SEG1 = 4Tq
				  //    |||+------- SEG1PH1
				  //    ||+-------- SEG1PH2
				  //    |+--------- SAM = 0 		-> one sample
				  //    +---------- SEG2PHTS = 1 	-> Use SEG2PH bits

			BRGCON3 = 0b01100010;
				  //    ||||||||
				  //    |||||||+--- <Unimplemented on PIC>
				  //    ||||||+---- WAKFIL = 1 		-> Use CAN filter on wakeup
				  //    |||||+----- <Unimplemented on PIC>
				  //    ||||+------ <Unimplemented on PIC>
				  //    |||+------- <Unimplemented on PIC>
				  //    ||+-------- SEG2PH0 = 011 	-> SEG2 = 4Tq
				  //    |+--------- SEG2PH1
				  //    +---------- SEG2PH2

			if (canBitRate != CAN_BITRATE_1MBPS)
				ThrowWarning(WARN_CAN_BITRATE);
			break;
	}
}





/*---------------------------------------------------------------------------\
| Function:		RequestCanMode
|
| Purpose:		Set the CAN mode bits and hang around until they change.
|
| Discussion:	WARNING! Hanging around until they change could be a really
|				long time... hundreds of ms to even seconds. EVENTUALLY, THIS
|				ROUTINE SHOULD HAVE A TIME OUT.
|
| Expects:		The wait to be ok
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

void RequestCanMode (enum CAN_MODES mode)
{
	// Request desired mode.

	CANCON = mode;

	// Wait till desired mode is set.

	while( (  CANSTAT & CAN_MODE_BITS) != mode );
}



/*---------------------------------------------------------------------------\
| Function:		CanSetMask
|
| Purpose:		Set the CAN mask for RB0 or RB1 with the specified mask value
|
| Discussion:	Passes the 16bit mask value for either the RB0 or RB1 mask to
|				the routine StoreCanMsgId which then updates all 11bits of the
|				mask.
|
| Expects:		-
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/


void CanSetMask(enum CAN_MASK canMaskNum,CanHeader_t mask )
{
	// Store the 11bit mask in the correct Receive Buffer
	// mask given by canMaskNum. Pass the pointer to the
	// correct register, assuming StoreCanMsgId realizes
	// the upper bits of the pointer are for the SFR bank.

	switch(canMaskNum)
	{

	    case CAN_RX_MASK_B0:
        StoreCanMsgId((uint8_t*)&RXM0SIDH, mask);
	        break;

	    case CAN_RX_MASK_B1:
        StoreCanMsgId((uint8_t*)&RXM1SIDH, mask);
	        break;

	    default:
	        ThrowWarning(WARN_CAN_MASK_NUM);
	        break;
    }
}





/*---------------------------------------------------------------------------\
| Function:		CanSetMask
|
| Purpose:		Set the CAN mask for RB0 or RB1 with the specified mask value
|
| Discussion:	Passes the 16bit mask value for either the RB0 or RB1 mask to
|				the routine StoreCanMsgId which then updates all 11bits of the
|				mask.
|
| Expects:		-
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

void CanSetFilter ( enum CAN_FILTER canFilterNum,CanHeader_t filter)
{
    switch(canFilterNum)
    {

    case CAN_RX_FILTER_B0_F0:
        StoreCanMsgId((uint8_t*)&RXF0SIDH, filter);
        break;

    case CAN_RX_FILTER_B0_F1:
        StoreCanMsgId((uint8_t*)&RXF1SIDH, filter);
        break;

    case CAN_RX_FILTER_B1_F2:
        StoreCanMsgId((uint8_t*)&RXF2SIDH, filter);
        break;

    case CAN_RX_FILTER_B1_F3:
        StoreCanMsgId((uint8_t*)&RXF3SIDH, filter);
        break;

    case CAN_RX_FILTER_B1_F4:
        StoreCanMsgId((uint8_t*)&RXF4SIDH, filter);
        break;

    case CAN_RX_FILTER_B1_F5:
        StoreCanMsgId((uint8_t*)&RXF5SIDH, filter);
        break;

    default:
        ThrowWarning(WARN_CAN_FILTER_NUM);
        break;
    }
}

/*---------------------------------------------------------------------------\
| Function:		StoreCanMsgId
|
| Purpose:		Store an 11bit value into 2 8bit registers
|
| Discussion:	Passes 16bit address of 8bit register to be updates - the
|				routine StoreCanMsgId then updates it and it's higher-in-
|				memory 8bit partner so that all 11bits of the mask are stored.
|
| Expects:		-
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

static void StoreCanMsgId(uint8_t* pRegister, CanHeader_t value )
{
	// Take bits 15:8 of the CanHeader_t and write them to the high byte
	// of the SFR register pair which holds a 11bits

    *pRegister =  value.bytes.high;

    // Point to the low byte (always one higher in the SFRs) of the
    // SFR register pair.

    pRegister++;

	// Take bits 7:5 of the CanHeader_t and write them to the low bute
	// of the SFR which also uses 7:5. And we can ignore 4:0 since
	// It's always extended (29bit) ID bits which we don't care about.

    *pRegister = value.bytes.low;

    GetCanMsgId(pRegister, &value);
}



/*---------------------------------------------------------------------------\
| Function:		GetCanMsgId
|
| Purpose:		Get an 11bit value packed in 2 8bit registers
|
| Discussion:	It's 4:00am. I feel like crap. How are you?
|
| Expects:		-
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

static void GetCanMsgId(uint8_t* pRegister, CanHeader_t *pValue)
{
	// Get bits 10:3 from the SFR register and store it in the high
	// byte of the CanHeader_t

	pValue->bytes.high = *pRegister;

	// Point to the next register of the pair (less signficant)

	pRegister++;

	// Get bits 2:0 from the LS SFR register and clear the rest of the
	// bits just in case.

	pValue->bytes.low = (*pRegister & 0b11100000);
}



/*---------------------------------------------------------------------------\
| Function:		SendCanMessage
|
| Purpose:		Load a CAN message into one of three CAN message queues.
|
| Discussion:	Chooses which queue to put the message in.
|				Disables CAN TX Buffer <n> empty interrupts for <N> cycles.
|				Sure we could encapsulate the message writes into a seperate
|				function calls but lets not waste cycles for the sake of
|				clearness ;)
|
| Expects:
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/30/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

void SendCanMessage (CanMessage_t* pMessage)
{

	static	uint8_t		i;
	static	uint16_t	id;


	// First, make sure we're not sending bad messages.

	if ( pMessage->header.bits.length > 8)
	{
		ThrowWarning(WARN_CAN_BAD_TX_DLC);
		return;
	}

	id = (pMessage->header.header & 0b1111111111100000); // FIXES WEIRD "0" CONSTANT PROBLEM.

	if ( (id < MIN_MESSAGE_ID) || (id > MAX_MESSAGE_ID) )
	{
		ThrowWarning(WARN_CAN_BAD_TX_ID);
		return;
	}

	// if Message's ID priority  >= high priority (i.e., it's LOWER)
	// then load message into canTx2Queue

	if ( id <= HIGH_PRIORITY_TX_CUTOFF )
	{
		DisableCanTx2Ints();

		// Copy over header info (ID, RTR, DLC)

		canTx2Queue[canTx2IndexIn].header.bytes.low = pMessage->header.bytes.low;
		canTx2Queue[canTx2IndexIn].header.bytes.high = pMessage->header.bytes.high;

		// Copy over all data bytes

		for (i = 0; i < pMessage->header.bits.length; ++i)
			canTx2Queue[canTx2IndexIn].data[i] = pMessage->data[i];

		// Increment the in index

		if ( canTx2IndexIn < (CAN_TX2_QUEUE_SIZE-1) )
			++canTx2IndexIn;
		else
			canTx2IndexIn = 0;

		// Check for overwrites. If we incremented the indexIn and it now
		// equals the indexOut then we've wrapped around the queue and
		// we have a queue overrun. In this case, don't write to the
		// queue - back off by one and throw a warning.

		if (canTx2IndexIn == canTx2IndexOut)
		{
			if (canTx2IndexIn == 0)
				canTx2IndexIn  = (CAN_TX2_QUEUE_SIZE-1);
			else
				--canTx2IndexIn;

			ThrowWarning(WARN_CAN_TX2_QUEUE_OVERFLOW);
		}

		// Whether or not the interrupts were enabled before, enable them now so that
		// we can load the TXB2 buffer right away.

		EnableCanTx2Ints();
	}

	// else if Message's ID priority >= low prioirty (i.e., it's LOWER)
	// then load message into canTx1Queue

	else if ( id <= NORMAL_PRIORITY_TX_CUTOFF )
	{
		DisableCanTx1Ints();

		// Copy over header info (ID, RTR, DLC)

		canTx1Queue[canTx1IndexIn].header.bytes.low = pMessage->header.bytes.low;
		canTx1Queue[canTx1IndexIn].header.bytes.high = pMessage->header.bytes.high;

		// Copy over all data bytes

		for (i = 0; i < pMessage->header.bits.length; ++i)
			canTx1Queue[canTx1IndexIn].data[i] = pMessage->data[i];

		// Increment the in index

		if ( canTx1IndexIn < (CAN_TX1_QUEUE_SIZE-1) )
			++canTx1IndexIn;
		else
			canTx1IndexIn = 0;

		// Check for overwrites. If we incremented the indexIn and it now
		// equals the indexOut then we've wrapped around the queue and
		// we have a queue overrun. In this case, don't write to the
		// queue - back off by one and throw a warning.

		if (canTx1IndexIn == canTx1IndexOut)
		{
			if (canTx1IndexIn == 0)
				canTx1IndexIn  = (CAN_TX1_QUEUE_SIZE-1);
			else
				--canTx1IndexIn;

			ThrowWarning(WARN_CAN_TX1_QUEUE_OVERFLOW);
		}

		// Whether or not the interrupts were enabled before, enable them now so that
		// we can load the TXB1 buffer right away.

		EnableCanTx1Ints();
	}

	// else load message into CanTx0Queue (must be a low prioirty message)

	else
	{
		DisableCanTx0Ints();

		// Copy over header info (ID, RTR, DLC)

		canTx0Queue[canTx0IndexIn].header.bytes.low = pMessage->header.bytes.low;
		canTx0Queue[canTx0IndexIn].header.bytes.high = pMessage->header.bytes.high;

		// Copy over all data bytes

		for (i = 0; i < pMessage->header.bits.length; ++i)
			canTx0Queue[canTx0IndexIn].data[i] = pMessage->data[i];

		// Increment the in index

		if ( canTx0IndexIn < (CAN_TX0_QUEUE_SIZE-1) )
			++canTx0IndexIn;
		else
			canTx0IndexIn = 0;

		// Check for overwrites. If we incremented the indexIn and it now
		// equals the indexOut then we've wrapped around the queue and
		// we have a queue overrun. In this case, don't write to the
		// queue - back off by one and throw a warning.

		if (canTx0IndexIn == canTx0IndexOut)
		{
			if (canTx0IndexIn == 0)
				canTx0IndexIn  = (CAN_TX0_QUEUE_SIZE-1);
			else
				--canTx0IndexIn;

			ThrowWarning(WARN_CAN_TX0_QUEUE_OVERFLOW);
		}

		// Whether or not the interrupts were enabled before, enable them now so that
		// we can load the TXB1 buffer right away.

		EnableCanTx0Ints();
	}
}



/*---------------------------------------------------------------------------\
| Function:		CanTx2Isr
|
| Purpose:		Grab messages from the CanTx2Queue and load them into TXB2.
|
| Discussion:
|
| Expects:
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/30/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

void CanTx2Isr (void)
{
	static uint8_t* pRegister;
	static uint8_t	i;


	// If we're here, someone has PROBABLY written to the canTx2Queue and enabled
	// interrupts (See SendCanMessage()). We can assume that
	// canTx2IndexOut points to one past the last message transmitted
	// (e.g., points to the next message to transmit).

	// If the in and out indices are equal, however, there aren't any more messages
	// in the transmit queue so we MUST turn off interrupts otherwise
	// we'll keep coming back here and thus lock up the PIC.

	if (TXB0REQ == 1)
	{
		#ifdef CANTXB2INT_HP
		ThrowWarning_HP(WARN_CAN_TXB2_NOT_EMPTY);
		#else
		ThrowWarning_LP(WARN_CAN_TXB2_NOT_EMPTY);
		#endif
		DisableCanTx2Ints();
	}

	else
	{
		if (canTx2IndexOut == canTx2IndexIn)
		{
			#ifdef CANTXB2INT_HP
			ThrowWarning_HP(WARN_CAN_TX2_QUEUE_EMPTY);
			#else
			ThrowWarning_LP(WARN_CAN_TX2_QUEUE_EMPTY);
			#endif
			DisableCanTx2Ints();
		}

		else
		{
			// Grab a message from canTx2Queue and stuff it into TXB2. Note that we're
			// not going to bother to setup CANCON to move the TXB2 registers into the
			// access bank - although for future renditions this could be a good idea.

			// Copy over the message ID

			TXB2SIDL = canTx2Queue[canTx2IndexOut].header.bytes.low & 0b1110000;
			TXB2SIDH = canTx2Queue[canTx2IndexOut].header.bytes.high;

			// Copy over the DLC and RTR bits

			TXB2DLC =  canTx2Queue[canTx2IndexOut].header.bytes.low & 0b00001111;
			TXB2RTR = canTx2Queue[canTx2IndexOut].header.bits.request;

			// Copy over all data bytes

			pRegister = &TXB2D0;

			for (i = 0; i < canTx2Queue[canTx2IndexOut].header.bits.length; ++i)
				*pRegister++ = canTx2Queue[canTx2IndexOut].data[i];

			// And send that puppy!

			TXB0REQ = 1;

			// 	Ok, it's out of our hands now. Increment the out index to the CIRCULAR queue.

			if (canTx2IndexOut < (CAN_TX2_QUEUE_SIZE-1))
				++canTx2IndexOut;
			else
				canTx2IndexOut = 0;

			// We've increased the counter, so check to see if the two pointers are equal.
			// If they are, then there's nothing in canTx2Queue to send to turn off transmit
			// buffer empty interrupts.

			if (canTx2IndexOut == canTx2IndexIn)
				DisableCanTx2Ints();
		}
	}
}



/*********************************************************************
 * Function:        bit CANSendMessage(unsigned long id,
 *                                  BYTE *Data,
 *                                  BYTE DataLen,
 *                                  enum CAN_TX_MSG_FLAGS MsgFlags)
 *
 * PreCondition:    None
 *
 * Input:           id          - CAN message identifier.
 *                                Only 11 or 29 bits may be used
 *                                depending on standard or extended
 *                                message type.
 *                  Data        - Data bytes of upto 8 bytes in length
 *                  DataLen     - Data length from 1 thru 8.
 *                  MsgFlags    - One or CAN_TX_MSG_FLAGS values ANDed
 *                                together
 *
 * Output:          If at least one empty transmit buffer is found,
 *                  given message is queued to be transmitted. If none
 *                  found FALSE value is returned.
 *
 * Side Effects:    None
 *
 ********************************************************************/
/*bit CANSendMessage(unsigned long id,
                     BYTE* Data,
                     BYTE DataLen,
                     enum CAN_TX_MSG_FLAGS MsgFlags)
{
    BYTE i;
    BYTE *ptr;

    // Find the first empty transmitter.
    if ( TXB0CON_TXREQ == 0 )
    {
        // TxBuffer0 is empty.  Set WIN bits to point to TXB0
        CANCON &= 0b11110001;
        CANCON |= 0b00001000;
    }
    else if ( TXB1CON_TXREQ == 0 )
    {
        // TxBuffer1 is empty. Set WIN bits to point to TXB1
        CANCON &= 0b11110001;
        CANCON |= 0b00000110;
    }
    else if ( TXB2CON_TXREQ == 0 )
    {
        // TxBuffer2 is empty. Set WIN bits to point to TXB2
        CANCON &= 0b11110001;
        CANCON |= 0b00000100;
    }
    else
        // None of the transmit buffers were empty.
        return FALSE;

    /*
     * Now that WIN has remapped RXB0 to empty buffer, simply
     * populate RXB0 buffer


    // Set transmit priority.
    RXB0CON = MsgFlags & CAN_TX_PRIORITY_BITS;

    // Populate Extended identifier information only if it is
    // desired.
    if ( !(MsgFlags & CAN_TX_FRAME_BIT) )
        CANIDToRegs((BYTE*)&RXB0SIDH, id, CAN_CONFIG_XTD_MSG);
    else
        CANIDToRegs((BYTE*)&RXB0SIDH, id, CAN_CONFIG_STD_MSG);

    RXB0DLC = DataLen;

    if ( !(MsgFlags & CAN_TX_RTR_BIT) )
        RXB0DLC |= 0b01000000;

    // Populate data values.
    ptr = (BYTE*)&RXB0D0;
    for ( i = 0; i < DataLen; i++ )
        ptr[i] = Data[i];

    /*
     * Mark this buffer as ready to start transmit.
     * We are not using C bit field structure because RXB0 registers
     * are remapped to one of the empty transmit buffers and their
     * bit3 is not same as RXB0CON bit3.  To avoid confusion, in-line
     * assembly is used to directly set bit 3 of corresponding TXBnCON
     * register.

#if defined(MCHP_C18)
     _asm
     bsf RXB0CON, 3, 0
    _endasm
#endif
#if defined(HITECH_C18)
       asm("bsf _RXB0CON,3");
#endif

    /*
     * Restore CAN buffer mapping so that subsequent access to RXB0
     * buffers are to the real RXB0 buffer.

    CANCON &= 0b11110001;

    return TRUE;
} */
