/*---------------------------------------------------------------------------\
| Project:			Generic LV2 PIC18F458 CAN node firmware
|
| File:				interrupts.c
|
| Description:		Interrupt Handlers.
|
| Global Routines:	InitializeInterrupts
|
| Local Routines:	None
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/01/2002 | Stolen from ADG code.
\---------------------------------------------------------------------------*/


//------------------------------------------------------------------------
//  #includes
//------------------------------------------------------------------------

#include <pic18.h>    	// Include for PICC-18
#include "pic18fxx8.h"	// Register definitions for the PIC18F458
#include "project.h"	// Includes for the project (Globals + misc)
#include "main.h"		// Includes for the mainline
#include "ints.h"		// Includes for interrupts
#include "excepts.h"	// Includes for exception handling
#include "ports.h"		// Includes for the port peripherals
#include "timer2.h"		// Includes for the Timer2 peripheral
#include "serial.h"		// Includes for the serial peripheral (UART )
#include "can.h"		// Includes for the CAN peripheral


//------------------------------------------------------------------------
//  Local function prototypes
//------------------------------------------------------------------------


/*---------------------------------------------------------------------------\
| Function:		HighPriorityISR
|
| Purpose:		Handle all of the high priority interrupts
|
| Discussion:	Checks to see what interrupted us. Why all this weird #ifdef
|				stuff? Because the PIC can't due recursive calls, and we have
|				an error function ("ThrowWarning") with a mainline, high
|				priority, and low priority flavors. In order for the compiler
|				to not yell at us, we need to figure out which interrupts
|				are which priority and #ifdef the correct error functions.
|				Ugly, but it's the best we can do (for now).
|
| Expects:		all interrupts to be disabled.
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 02/10/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

// DECLARE THIS FUNCTION TO BE THE HIGH PRIOIRTY INTERRUPT SERVICE ROUTINE
// USING THE "interrupt" KEYWORD.


void interrupt HighPriorityISR (void)
{
	static bit hpTest;

	hpTest = FALSE;

	#ifdef	CANTXB2INT_HP
	if ( CanTxb2InterruptPending() == TRUE )
	{
		CanTx2Isr();
		hpTest = TRUE;
	}
	#endif

	#ifdef	TIMER2INT_HP
	if ( Timer2InterruptPending() == TRUE )
	{
		Timer2Isr();
		hpTest = TRUE;
	}
	#endif

	#ifdef	SERIALTXINT_HP
	if ( SerialTxInterruptPending() == TRUE )
	{
		SerialTxIsr();
		hpTest = TRUE;
	}
	#endif

	#ifdef	SERIALRXINT_HP
	if ( SerialRxInterruptPending() == TRUE )
	{
		SerialRxIsr();
		hpTest = TRUE;
	}
	#endif

	// Throw a warning if we don't know why were interrupted.

	if (hpTest == FALSE)
		ThrowWarning_HP(WARN_ISR_UNKNOWN_HPI);
}



/*---------------------------------------------------------------------------\
| Function:		LowPriorityISR
|
| Purpose:		Handle all of the low prioirty interrupts
|
| Discussion:	Checks to see what interrupted us.
|
| Expects:		Low prioirty interrupts disabled but high prioirty may still
|				be enabled.
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 02/10/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

// DECLARE THIS FUNCTION TO BE THE LOW PRIOIRTY INTERRUPT SERVICE ROUTINE
// USING THE "interrupt low-priority" KEYWORDS.

void interrupt low_priority LowPriorityISR (void)
{
	static bit lpTest;

	lpTest = FALSE;

	#ifdef	CANTXB2INT_LP
	if CanTxb2InterruptPending() == TRUE )
	{
		CanTx2Isr();
		lpTest = TRUE;
	}
	#endif

	#ifdef	TIMER2INT_LP
	if ( Timer2InterruptPending() == TRUE )
	{
		Timer2Isr();
		lpTest = TRUE;
	}
	#endif

	#ifdef	SERIALTXINT_LP
	if ( SerialTxInterruptPending() == TRUE )
	{
		SerialTxIsr();
		lpTest = TRUE;
	}
	#endif

	#ifdef	SERIALRXINT_LP
	if ( SerialRxInterruptPending() == TRUE )
	{
		SerialRxIsr();
		lpTest = TRUE;
	}
	#endif

	// Throw a warning if we don't know why were interrupted.

	if (lpTest == FALSE)
		ThrowWarning_LP(WARN_ISR_UNKNOWN_LPI);
}





/*---------------------------------------------------------------------------\
| Function:		InitializeInterrupts
|
| Purpose:		Initialize the PIC's interrupts
|
| Discussion:	Huh?
|
| Expects:		All interupts to be off (after all, we were just reseted).
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

void InitializeInterrupts (void)
{
	// Turn on "high and low" priorities, not old peripheral/global

	EnablePrioritizedInts();

	// NOTE: All interrupts can be either low or high prioirity. Which one
	// they are is defined in "ints.h" by #defines. We still need to set the
	// correct bits, which we do here. Why the funny #ifdef? The lack of
	// recursion in PICs makes this best to do in pre-processor land. See
	// "ints.h" for more info.

	#ifdef CANTXB2INT_HP
	CanTxb2PrioritySetting() = HIGHPRIORITY;
	#else
	Timer2PrioritySetting() = LOWPRIORITY;
	#endif

	#ifdef TIMER2INT_HP
	Timer2PrioritySetting() = HIGHPRIORITY;
	#else
	Timer2PrioritySetting() = LOWPRIORITY;
	#endif

	#ifdef SERIALTXINT_HP
	SerialTxPrioritySetting() = HIGHPRIORITY;
	#else
	SerialTxPrioritySetting() = LOWPRIORITY;
	#endif

	#ifdef SERIALRXINT_HP
	SerialRxPrioritySetting() = HIGHPRIORITY;
	#else
	SerialRxPrioritySetting() = LOWPRIORITY;
	#endif


	// Turn on ints and away we go!

	EnableHighPriorityInts();
	EnableLowPriorityInts();
}

