/*---------------------------------------------------------------------------\
| Project:			Generic LV2 PIC18F458 CAN node firmware
|
| File:				serial.c
|
| Description:		Asynchronous serial port (UART) routines
|
| Global Routines:	InitializeSerial()
|					SerialRxIsr()
|					SerialTxIsr()
|					SendSerialByte()
|					GetSerialByte()
|
| Local Routines:	None
|
|
| Overview:
|
|  Initialization:
|
|	Call InitializeSerial to set everything up. See the enums for more details.
|
|  Transmission:
|
|	Mainline calls SendSerialByte which stuffs a byte into the SerialTxQueue.
|	SerialTxIsr then grabs bytes from the queue on a UART transmit buffer
|	interrupt and stuffs them into the UART transmit register (TXREG).
|	In other words:
|
|	Main -> SendSerialByte -> SerialTxQueue -> SerialTxISR -> TXREG
|
|  Reception:
|
|	SerialRxIsr grabs bytes from the UART Receive register (RXREG) on a
|	Receiver Buffer Full Interrupt and stuffs them into the SerialRxQueue.
|	Mainline routines can then grab bytes from the queue using GetSerialByte.
|	In other words:
|
|	RCREG -> SerialRxISR -> SerialRxQueue -> GetSerialByte -> Main
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/01/2002 | Stolen from ADG code.
\---------------------------------------------------------------------------*/

//------------------------------------------------------------------------
//  #includes
//------------------------------------------------------------------------

#include <p18f458.h>    // Register definitions for the PIC18F458
#include "project.h"	// Includes for the project (misc)
#include "ints.h"		// Includes for interrupts
#include "serial.h"		// Includes for the serial peripheral
#include "excepts.h"	// Includes for exception handling


//------------------------------------------------------------------------
//  Module-wide variables
//------------------------------------------------------------------------

// FIFO circular queue and indices for data being transmitted over the UART

uint8_t	indexInSerialTxQueue;
uint8_t	indexOutSerialTxQueue;
uint8_t	serialTxQueue[SERIALTXQUEUELENGTH];

// FIFO circular queue and indices for data being received over the UART

uint8_t	indexInSerialRxQueue;
uint8_t	indexOutSerialRxQueue;
uint8_t	serialRxQueue[SERIALRXQUEUELENGTH];


BOOL	gSerialRxDataAvailable;

/*---------------------------------------------------------------------------\
| Function:		InitializeSerial
|
| Purpose:		Initialize the UART peripheral
|
| Discussion:	Initialize the UART peripheral, and do as little damage as
|				possible while doing it.
|
| Expects:		RCSTA, TXSTA to be in the known RESET state.
|
| Arguments:	txEnable: 	0 = transmitter off, 1 = transmitter on
|				txIntEnable:INT_ENABLE_TYPE: disabled, low, or high prioirty
|				rxEnable: 	0 = receiver off, 1 = receiver on
|				rxIntEnable:INT_ENABLE_TYPE: disabled, low, or high prioirty
|				bitRate: 	SERIAL_BITRATE type: 4800, 9600 or 19200bps
|
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/01/2002 | Stolen from ADG code.
\---------------------------------------------------------------------------*/

void InitializeSerial (enum SERIAL_INIT_MASK optionsMask)
{
	// Coming out of reset, we can rely on the following "RESET" values:
	//
	//	RCSTA = 0b00000000;
	//			  ||||||||
	//			  |||||||+----------- RX9D  = u = 9th bit (unknown)
	//			  ||||||+------------ OERR  = 0 = no overrun error
	//			  |||||+------------- FERR  = 0 = no frame error
	//			  ||||+-------------- ADDEN = 0 = 9th bit adrs detection off
	//			  |||+--------------- CREN  = 0 = continuous receive off
	//			  ||+---------------- SREN  = 0 = single receive off
	//			  |+----------------- RX9   = 0 = 8bit reception
	//			  +------------------ SPEN  = 0 = serial port disabled
	//
	//	TXSTA = 0b00000010;
	//			  ||||||||
	//			  |||||||+----------- TX9D = 0 = 9th bit to transmit
	//			  ||||||+------------ TRMT = 1 = transmit shift regst empty
	//			  |||||+------------- BRGH = 0 = bitrate clock @ low speed
	//			  ||||+-------------- <Unimplemented on PIC>
	//			  |||+--------------- SYNC = 0 = asynchronous mode
	//			  ||+---------------- TXEN = 0 = transmissions off
	//			  |+----------------- TX9D = 0 = 8bit transmissions
	//			  +------------------ CSRC = 0 = sync. clock source
	//
	//	SPBRG = 0				// Does this mean the divider is off?
	//
	//  PIE1bits.TXIE = 0;		// DISABLE UART TX interrupt
	//	PIE1bits.RXIE = 0;		// DISABLE UART RX interrupt
	//	IPR1bits.TXIP = 0;		// UART TX interrupt = low priority
	//	IPR1bits.RCIP = 0;		// UART RX interrupt = low priority


	// So now, based on these reset values, we can apply what changes
	// we want.

	// Select a bit rate of 4800bps, 9600bps or 19.2kbps.

	switch (optionsMask & 0b00000011)
	{

		case SERIAL_BITRATE_4800:
			SPBRG = 129; 			// ASSUMES FOSC = 40MHZ
			TXSTAbits.BRGH = 0;		// ASSUMES FOSC = 40MHZ
			break;

		case SERIAL_BITRATE_9600:
			SPBRG = 64;  			// ASSUMES FOSC = 40MHZ
			TXSTAbits.BRGH = 0;		// ASSUMES FOSC = 40MHZ
			break;

		case SERIAL_BITRATE_19200:
			SPBRG = 129; 			// ASSUMES FOSC = 40MHZ
			TXSTAbits.BRGH = 1;		// ASSUMES FOSC = 40MHZ
			break;

		default:					// Default: 9600bps
			SPBRG = 64;  			// ASSUMES FOSC = 40MHZ
			TXSTAbits.BRGH = 0;		// ASSUMES FOSC = 40MHZ
			throwWarning(WARN_SERIAL_BITRATE);
			break;
	}

	// Select transmitter enable, interrupt enable, and interrupt priority

	if (optionsMask & SERIAL_TRANSMIT_ENABLE)
	{

		RCSTAbits.SPEN = 1;
		TXSTAbits.TXEN = 1;

		if (optionsMask & SERIAL_TRANSMIT_ENABLE_INTS)
		{
			EnableSerialTxInts();
			if (optionsMask & SERIAL_TRANSMIT_INTS_HIGHPRIORITY)
				SetHPSerialTxInts();
			else
				SetLPSerialTxInts();
		}
	}

	// Select receiver enable, interrupt enable, and interrupt priority

	if (optionsMask & SERIAL_RECEIVE_ENABLE)
	{

		RCSTAbits.SPEN = 1;
		RCSTAbits.CREN = 1;

		if (optionsMask & SERIAL_RECEIVE_ENABLE_INTS)
		{
			EnableSerialRxInts();
			if (optionsMask & SERIAL_RECEIVE_INTS_HIGHPRIORITY)
				SetHPSerialRxInts();
			else
				SetLPSerialRxInts();
		}
	}
}




/*---------------------------------------------------------------------------\
| Function:		SerialTxIsr
|
| Purpose:		Handle the UART transmit interrupt by stuffing the transmit
|				register.
|
| Discussion:	Grab bytes from the FIFO circular queue and stuff them
|				into the UART
|
| Expects:		LOW PRIORITY INTERRUPTS TO BE DISABLED.
|				NO READS OR WRITES TO THE QUEUE.
|				NO CHANGE IN THE UART TX STATE.
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/01/2002 | Stolen from ADG code.
\---------------------------------------------------------------------------*/

void SerialTxIsr (void)
{
	// If we're here, someone has written to the serialTxQueue and enabled
	// interrupts (See SerialTransmit()). We can assume that
	// indexOutSerialTxQueue points to one past the last byte transmitted
	// (e.g., the next byte to transmit).

	// Grab from the cicular queue and transmit it.

	TransmitAndClearInterupt(serialTxQueue[indexOutSerialTxQueue]);

	// 	Increment the index to the CIRCULAR queue.

	if (indexOutSerialTxQueue < (SERIALTXQUEUELENGTH-1))
		++indexOutSerialTxQueue;
	else
		indexOutSerialTxQueue = 0;

	// If the in and out indices are equal, we're done transmitting
	// and we can turn off transmit interrupts.

	if (indexOutSerialTxQueue == indexInSerialTxQueue)
		DisableSerialTxInts();
}




/*---------------------------------------------------------------------------\
| Function:		SerialRxIsr
|
| Purpose:		Handle the UART Receive interrupt
|
| Discussion:	Receive stuff, possibly at 1920 Bytes per second (19.2kbps)
|
| Expects:		-
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/01/2002 | Stolen from ADG code.
\---------------------------------------------------------------------------*/

void SerialRxIsr (void)
{

	// Read in the received byte, clear interrupt

	serialRxQueue[indexInSerialRxQueue] = ReceiveAndClearInterrupt();

	// 	Increment the index to the CIRCULAR queue.

	if (indexInSerialRxQueue < (SERIALRXQUEUELENGTH-1))
		++indexInSerialRxQueue;
	else
		indexInSerialRxQueue = 0;

	// Check for overwrites. If we incremented the indexIn and it now
	// equals the indexOut then we've wrapped around the queue and
	// we have a queue overrun. In this case, don't write to the
	// queue - back off by one and throw a warning.

	if (indexInSerialRxQueue == indexOutSerialRxQueue)
	{
		if (indexInSerialRxQueue == 0)
			indexInSerialRxQueue  = (SERIALRXQUEUELENGTH-1);
		else
			--indexInSerialRxQueue;

		throwWarning(SERIAL_RXQUEUE_OVERRUN);
	}

	// In any case, let everyone know that there's data to be had in the queue

	gSerialRxDataAvailable = 1;
}




/*---------------------------------------------------------------------------\
| Function:		SendSerialByte
|
| Purpose:		Stuff the FIFO circular queue for the UART TX.
|
| Discussion:	Inserts a byte into the buffer.
|				Disables UART TX Buffer Empty interrupts for <N> cycles.
|
| Expects:		Nothing.
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/01/2002 | Stolen from ADG code.
\---------------------------------------------------------------------------*/

void SendSerialByte (uint8_t byte)
{
	// Disable Serial TX interrupts for a few cycles while we stuff
	// the queue.

	DisableSerialTxInts();

	// Store the byte at the current index.

	serialTxQueue[indexInSerialTxQueue] = byte;

	// 	Increment the index to the CIRCULAR queue.

	if (indexInSerialTxQueue < (SERIALTXQUEUELENGTH-1))
		++indexInSerialTxQueue;
	else
		indexInSerialTxQueue = 0;

	// Check for overwrites. If we incremented the indexIn and it now
	// equals the indexOut then we've wrapped around the queue and
	// we have a queue overrun. In this case, don't write to the
	// queue - back off by one and throw a warning.

	if (indexInSerialTxQueue == indexOutSerialTxQueue)
	{
		if (indexInSerialTxQueue == 0)
			indexInSerialTxQueue  = (SERIALTXQUEUELENGTH-1);
		else
			--indexInSerialTxQueue;

		throwWarning(SERIAL_TXQUEUE_OVERRUN);
	}

	// In any case, enable the transmit interrupts to transmit date from
	// the queue.

	EnableSerialTxInts();
}





/*---------------------------------------------------------------------------\
| Function:		GetSerialByte
|
| Purpose:		Get a byte from the FIFO circular queue for the UART RX.
|
| Discussion:	Gets a byte from the buffer.
|				Disables UART RX Buffer Full interrupts for <N> cycles.
|
| Expects:		Nothing.
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/01/2002 | Stolen from ADG code.
\---------------------------------------------------------------------------*/

uint8_t GetSerialByte (void)
{
	uint8_t temp;

	// read the queue

	temp = serialRxQueue[indexOutSerialRxQueue];

	// 	Increment the index to the CIRCULAR queue.

	if (indexOutSerialRxQueue < (SERIALRXQUEUELENGTH-1))
		++indexOutSerialRxQueue;
	else
		indexOutSerialRxQueue = 0;

	// Temporarily disable the serial receive interrupts

	DisableSerialRxInts();

	// If the in and out indices are equal, we're done transmitting
	// and we can turn off transmit interrupts.

	if (indexOutSerialRxQueue == indexInSerialRxQueue)
		gSerialRxDataAvailable = 0;

	// Re-enable receive interrupts

	EnableSerialRxInts();

	return(temp);
}


